use strict;
use warnings;
use Time::Piece;
use Text::Xslate;
use Path::Tiny 'path';
# Return two closures to make SQL interaction easier.
# $sql is for SELECT queries.
# $do  is for INSERT, UPDATE, and DELETE statements and anything mutatey.
sub sql_fn {
  my $dbh = shift;
  my $sql = sub {
    my $query = shift;
    $dbh->selectall_array($query, { Slice => {} }, @_);
  };
  my $do = sub {
    my $statement = shift;
    $dbh->do($statement, { Slice => {} }, @_);
  };
  ($sql, $do);
}

# Remove whitespace from beginning and end of string.
sub trim {
  my $s = shift;
  $s =~ s/^\s*//;
  $s =~ s/\s*$//;
  return $s;
}

# YouTube link for video_id
sub yt {
  my $id = shift;
  "https://youtube.com/watch?v=$id";
}

# Thumbnail link for video_id
sub thumb {
  my $id = shift;
  my $x = substr $id, 0, 2;
  "https://site.weathernews.jp/app/wnlive/thumb/$x/$id.jpg"
}

# mobile video page for aupay video_id
sub aupay_vid {
  my $id = shift;
  "https://wowma.jp/event/live-tv/play.html?live_id=$id";
}

# m3u8 link for aupay video_id
sub aupay_m3u8 {
  my $id = shift;
  my $quality = shift || '1080p';
  "https://vod.check.tv/lives/$id/$quality/index.m3u8";
}

# thumbnail for aupay video_id
sub aupay_th {
  my $id = shift;
  "https://image.check.tv/lives/$id/vertical.jpg";
}

# remove the T so that the datetime wraps better when on a narrow screen.
sub date_time {
  my $dt = shift;
  $dt =~ s/T/ /;
  return $dt;
}

my $_dir = path(__FILE__)->parent;
my $tx = Text::Xslate->new(
  path     => [ $_dir ],
  header   => [ "./xslate/_header.html" ],
  footer   => [ "./xslate/_footer.html" ],
  function => {
    yt         => \&yt,
    th         => \&thumb,
    aupay_vid  => \&aupay_vid,
    aupay_m3u8 => \&aupay_m3u8,
    aupay_th   => \&aupay_th,
    date_time  => \&date_time,
  }
);
sub page {
  my $name = shift;
  my $vars = shift;
  my $template = "./xslate/$name.html";
  return $tx->render($template, $vars);
}

sub sum {
  my $total = 0;
  $total += $_ for (@_);
  return $total;
}

sub dist_to_arrayrefs {
  my $rows = shift;
  my $res = { headers => [], rows => [] };
  return $res unless (scalar @$rows);
  my ($main) = grep { $_ ne "appearances" } keys %{$rows->[0]};
  $res->{headers} = [$main, "appearances"];
  $res->{rows} = [ map { [ $_->{$main}, $_->{appearances} ] } @$rows ];
  $res->{total} = sum(map { $_->{appearances} } @$rows);
  return $res;
}

my $H3 = (60 * 60 * 3); # 3 hours
my $H9 = (60 * 60 * 9); # 9 hours in seconds
# return true if $jst should be broadcasting live $now
sub is_live {
  my $jst = shift;
  my $now = shift || localtime();
  # take the jst and subtract 9 hours from it to normalize it to utc
  my $utc = Time::Piece->strptime($jst, "%Y-%m-%dT%H:%M:%S+09:00") - $H9;
  # compare $now - $utc
  my $diff = $now - $utc;
  my $duration = $H3;
  # Most shows are 3 hours, but 23 and 00 are 1 and 5 hours.
  $duration = (60 * 60 * 1) if ($utc->hour == 14); # 23:00 JST
  $duration = (60 * 60 * 5) if ($utc->hour == 15); # 00:00 JST
  # a value between 0 and $duration is good
  return ($diff >= 0 && $diff <= $duration);
}

my $LIVE_SEARCH_LIMIT = 22; # unlikely to find live shows beyond this
sub mark_live {
  my $rows = shift;
  my $i = 0;
  for (@$rows) {
    #warn "$i $_->{caster}\n";
    if (is_live($_->{jst})) {
      $_->{live} = " live";
    }
    $i++;
    last if ($i >= $LIVE_SEARCH_LIMIT);
  }
}

1;
